#!/usr/bin/env python3
"""
vtree - A modern, interactive terminal-based file tree viewer with file panel support
"""

import os
import sys
import subprocess
from pathlib import Path
from typing import List, Optional, Set
from dataclasses import dataclass
from datetime import datetime

from textual.app import App, ComposeResult
from textual.containers import Horizontal, Vertical, Container
from textual.widgets import Tree, Footer, Header, Static, Label, DataTable
from textual.binding import Binding
from textual.reactive import reactive
from textual.widget import Widget
from textual.message import Message
from rich.text import Text
import click


@dataclass
class FileNode:
    path: Path
    is_dir: bool
    size: Optional[int] = None
    children: Optional[List['FileNode']] = None


class FileTree(Tree):
    """Custom Tree widget for file system navigation."""
    
    def __init__(self, root_path: Path, show_hidden: bool = False, show_files_inline: bool = True, *args, **kwargs):
        # Initialize with the root path name as the tree label
        super().__init__(f"[bold blue]{root_path.name}[/bold blue]", *args, **kwargs)
        self.root_path = root_path
        self.show_hidden = show_hidden
        self.show_files_inline = show_files_inline
        self.ignore_patterns = {
            '.git', '__pycache__', '.pytest_cache', 'node_modules',
            '.venv', 'venv', '.DS_Store', '.idea', '.vscode',
            'dist', 'build', '.egg-info'
        }
        self.populate_tree()
    
    def should_ignore(self, path: Path) -> bool:
        """Check if a file/directory should be ignored"""
        name = path.name
        
        # Hidden files/dirs (starting with .)
        if name.startswith('.') and not self.show_hidden:
            if name not in {'.gitignore', '.env.example', '.github'}:
                return True
        
        # Check ignore patterns
        if any(pattern in name for pattern in self.ignore_patterns):
            return True
            
        return False
    
    def get_file_size(self, path: Path) -> Optional[int]:
        """Get file size safely"""
        try:
            return path.stat().st_size if path.is_file() else None
        except (OSError, PermissionError):
            return None
    
    def format_size(self, size: Optional[int]) -> str:
        """Format file size in human readable format"""
        if size is None:
            return ""
        
        units = ['B', 'KB', 'MB', 'GB', 'TB']
        size_float = float(size)
        unit_index = 0
        
        while size_float >= 1024 and unit_index < len(units) - 1:
            size_float /= 1024
            unit_index += 1
        
        if unit_index == 0:
            return f" [dim]({int(size_float)} {units[unit_index]})[/dim]"
        else:
            return f" [dim]({size_float:.1f} {units[unit_index]})[/dim]"
    
    
    def populate_tree(self):
        """Populate the tree with file system data"""
        self.clear()
        
        # Set data on root node
        self.root.data = {"path": self.root_path, "type": "dir"}
        
        # Add children directly to root
        self._add_children(self.root, self.root_path)
        
        # Force expand the root node immediately
        self.root.expand()
        self.root.expanded = True
    
    def _add_children(self, parent_node, parent_path: Path, max_depth: int = 5, current_depth: int = 0):
        """Recursively add children to tree nodes"""
        if current_depth >= max_depth:
            return
        
        try:
            # Get all items and sort: directories first, then files, both alphabetically
            items = list(parent_path.iterdir())
            items.sort(key=lambda x: (not x.is_dir(), x.name.lower()))
            
            for item in items:
                if self.should_ignore(item):
                    continue
                
                name = item.name
                
                if item.is_dir():
                    label = f"[bold yellow]{name}[/bold yellow]"
                    node_data = {"path": item, "type": "dir"}
                    child_node = parent_node.add(label, data=node_data)
                    
                    # Add children recursively
                    self._add_children(child_node, item, max_depth, current_depth + 1)
                elif self.show_files_inline:
                    size_str = self.format_size(self.get_file_size(item))
                    label = f"{name}{size_str}"
                    node_data = {"path": item, "type": "file", "size": self.get_file_size(item)}
                    parent_node.add_leaf(label, data=node_data)
                    
        except PermissionError:
            parent_node.add_leaf("❌ [red]Permission denied[/red]")
    
    def refresh_tree(self, show_hidden: bool = None, show_files_inline: bool = None):
        """Refresh the tree display"""
        if show_hidden is not None:
            self.show_hidden = show_hidden
        if show_files_inline is not None:
            self.show_files_inline = show_files_inline
        self.populate_tree()


class InfoPanel(Static):
    """Information panel showing details about selected file/directory"""
    
    def __init__(self, *args, **kwargs):
        super().__init__("", *args, **kwargs)
        self.border_title = "Info"
    
    def update_info(self, path: Path, file_type: str, size: Optional[int] = None):
        """Update the info panel with file/directory information"""
        info_text = Text()
        
        # More compact format
        info_text.append(f"📍 {path.name}", style="bold cyan")
        
        if file_type == "file" and size is not None:
            size_str = self._format_size(size)
            info_text.append(f" ({size_str})", style="dim")
        
        info_text.append(f" • {file_type.title()}", style="yellow")
        info_text.append(f" • {path.parent}", style="dim")
        
        self.update(info_text)
    
    def _format_size(self, size: int) -> str:
        """Format file size in human readable format"""
        units = ['B', 'KB', 'MB', 'GB', 'TB']
        size_float = float(size)
        unit_index = 0
        
        while size_float >= 1024 and unit_index < len(units) - 1:
            size_float /= 1024
            unit_index += 1
        
        if unit_index == 0:
            return f"{int(size_float)} {units[unit_index]}"
        else:
            return f"{size_float:.1f} {units[unit_index]}"


class FileListPanel(DataTable):
    """Panel showing files in the selected directory"""
    
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.border_title = "Files"
        self.zebra_stripes = True
        self.cursor_type = "row"
        self.add_columns("Name", "Size", "Modified")
    
    def update_files(self, directory_path: Path):
        """Update the file list with files from the given directory"""
        self.clear()
        
        if not directory_path.is_dir():
            return
        
        try:
            items = []
            for item in directory_path.iterdir():
                if item.is_file():
                    try:
                        stat = item.stat()
                        size = stat.st_size
                        modified = datetime.fromtimestamp(stat.st_mtime)
                        items.append((item.name, size, modified))
                    except (OSError, PermissionError):
                        items.append((item.name, 0, datetime.now()))
            
            # Sort by name
            items.sort(key=lambda x: x[0].lower())
            
            for name, size, modified in items:
                size_str = self._format_size(size)
                date_str = modified.strftime("%Y-%m-%d %H:%M")
                self.add_row(name, size_str, date_str)
                
        except PermissionError:
            self.add_row("Permission denied", "", "")
    
    def _format_size(self, size: int) -> str:
        """Format file size in human readable format"""
        units = ['B', 'KB', 'MB', 'GB', 'TB']
        size_float = float(size)
        unit_index = 0
        
        while size_float >= 1024 and unit_index < len(units) - 1:
            size_float /= 1024
            unit_index += 1
        
        if unit_index == 0:
            return f"{int(size_float)} {units[unit_index]}"
        else:
            return f"{size_float:.1f} {units[unit_index]}"


class HelpPanel(Static):
    """Help panel showing keyboard shortcuts"""
    
    def __init__(self, *args, **kwargs):
        super().__init__("", *args, **kwargs)
        self.border_title = "Help"
        
        help_text = Text()
        
        shortcuts = [
            ("q/Ctrl+C", "Quit"),
            ("r", "Refresh"),
            ("f", "Hidden files"),
            ("p", "File panel"),
            ("c", "Copy path"),
            ("h/?", "Help"),
            ("↑↓←→", "Navigate"),
            ("Mouse", "Click"),
        ]
        
        for i, (key, desc) in enumerate(shortcuts):
            if i > 0:
                help_text.append(" • ")
            help_text.append(f"{key}", style="bold cyan")
            help_text.append(f": {desc}", style="dim")
        
        self.update(help_text)


class VTreeApp(App):
    """A Textual app for displaying project trees."""
    
    CSS = """
    Screen {
        background: #1e1e1e;
        color: #d4d4d4;
    }
    
    #tree-container {
        height: 1fr;
        border: solid #3e3e3e;
        border-title-color: #808080;
    }
    
    #file-list-container {
        height: 1fr;
        max-height: 50%;
        border: solid #3e3e3e;
        border-title-color: #808080;
        margin-top: 1;
        background: #252526;
    }
    
    #info-container {
        height: auto;
        max-height: 8;
        border: solid #3e3e3e;
        border-title-color: #808080;
        margin-top: 1;
        background: #252526;
    }
    
    .hidden {
        display: none;
    }
    
    FileTree {
        height: 100%;
        background: #252526;
        color: #d4d4d4;
        scrollbar-color: #424242;
        scrollbar-color-hover: #4ec9b0;
    }
    
    InfoPanel {
        height: auto;
        padding: 1;
        background: #252526;
        color: #d4d4d4;
    }
    
    HelpPanel {
        height: auto;
        padding: 1;
        background: #252526;
        color: #808080;
    }
    
    FileListPanel {
        height: 100%;
        background: #252526;
        color: #d4d4d4;
    }
    
    DataTable > .datatable--header {
        background: #2d2d30;
        color: #cccccc;
    }
    
    DataTable > .datatable--cursor {
        background: #094771;
        color: #4ec9b0;
    }
    
    Tree > .tree--cursor {
        background: #094771;
        color: #4ec9b0;
    }
    
    Tree > .tree--highlight {
        background: #2a2d2e;
    }
    
    Header {
        background: #2d2d30;
        color: #cccccc;
    }
    
    Footer {
        background: #2d2d30;
        color: #808080;
    }
    """

    BINDINGS = [
        Binding("q", "quit", "Quit"),
        Binding("r", "refresh", "Refresh"),
        Binding("h,question_mark", "toggle_help", "Help"),
        Binding("ctrl+c", "quit", "Quit", show=False),
        Binding("f", "toggle_hidden", "Hidden Files"),
        Binding("p", "toggle_file_panel", "File Panel"),
        Binding("c", "copy_path", "Copy Path"),
    ]

    show_help = reactive(False)
    show_hidden = reactive(False)
    show_file_panel = reactive(False)
    current_selection = reactive(None)

    def __init__(self, root_path: str = "."):
        super().__init__()
        self.root_path = Path(root_path).resolve()
        self.title = f"vtree - {self.root_path.name}"
        self.sub_title = str(self.root_path)

    def compose(self) -> ComposeResult:
        """Create child widgets for the app."""
        yield Header()
        
        with Vertical():
            tree_container = Container(id="tree-container")
            tree_container.border_title = "Folders"
            tree_container.add_class("tree-container")
            with tree_container:
                yield FileTree(self.root_path, self.show_hidden, show_files_inline=not self.show_file_panel, id="file-tree")
            
            file_list_container = Container(id="file-list-container", classes="hidden" if not self.show_file_panel else "")
            file_list_container.border_title = "Files"
            with file_list_container:
                yield FileListPanel(id="file-list-panel")
            
            with Container(id="info-container"):
                yield InfoPanel(id="info-panel")
                yield HelpPanel(id="help-panel", classes="hidden" if not self.show_help else "")
        
        yield Footer()

    def on_mount(self) -> None:
        """Called when app starts."""
        tree = self.query_one(FileTree)
        tree.focus()
        
        # The root folder should already be expanded from populate_tree,
        # but also handle file panel mode if active
        if self.show_file_panel:
            file_list_panel = self.query_one(FileListPanel)
            file_list_panel.update_files(self.root_path)

    def on_tree_node_selected(self, event: Tree.NodeSelected) -> None:
        """Handle tree node selection."""
        if event.node.data:
            path = event.node.data["path"]
            file_type = event.node.data["type"]
            size = event.node.data.get("size")
            
            # Store current selection for copy functionality
            self.current_selection = {"path": path, "type": file_type}
            
            info_panel = self.query_one(InfoPanel)
            info_panel.update_info(path, file_type, size)
            
            # Update file list panel if it's visible and we selected a directory
            if self.show_file_panel and file_type == "dir":
                file_list_panel = self.query_one(FileListPanel)
                file_list_panel.update_files(path)

    def action_refresh(self) -> None:
        """Refresh the file tree."""
        tree = self.query_one(FileTree)
        tree.refresh_tree(self.show_hidden)

    def action_toggle_help(self) -> None:
        """Toggle the help panel."""
        self.show_help = not self.show_help
        help_panel = self.query_one(HelpPanel)
        
        if self.show_help:
            help_panel.remove_class("hidden")
        else:
            help_panel.add_class("hidden")

    def action_toggle_hidden(self) -> None:
        """Toggle showing hidden files."""
        self.show_hidden = not self.show_hidden
        tree = self.query_one(FileTree)
        tree.refresh_tree(self.show_hidden)

    def action_toggle_file_panel(self) -> None:
        """Toggle the file list panel."""
        self.show_file_panel = not self.show_file_panel
        file_list_container = self.query_one("#file-list-container")
        tree = self.query_one(FileTree)
        
        if self.show_file_panel:
            file_list_container.remove_class("hidden")
            tree.refresh_tree(show_files_inline=False)
            # Also update the file list panel with root directory files
            file_list_panel = self.query_one(FileListPanel)
            file_list_panel.update_files(self.root_path)
        else:
            file_list_container.add_class("hidden")
            tree.refresh_tree(show_files_inline=True)

    def action_copy_path(self) -> None:
        """Copy the current path to clipboard."""
        if not self.current_selection:
            return
        
        path = self.current_selection["path"]
        file_type = self.current_selection["type"]
        
        # For files, copy the parent directory path
        # For folders, copy the folder path itself
        if file_type == "file":
            copy_path = str(path.parent)
        else:
            copy_path = str(path)
        
        try:
            # Try to copy to clipboard using pbcopy (macOS) or xclip (Linux)
            if sys.platform == "darwin":
                subprocess.run(["pbcopy"], input=copy_path, text=True, check=True)
            elif sys.platform.startswith("linux"):
                subprocess.run(["xclip", "-selection", "clipboard"], input=copy_path, text=True, check=True)
            else:
                # Fallback: print the path for manual copying
                self.bell()
                return
            
            # Visual feedback - update info panel with copied path message
            info_panel = self.query_one(InfoPanel)
            from rich.text import Text
            feedback_text = Text()
            feedback_text.append("Path copied to clipboard: ", style="bold green")
            feedback_text.append(copy_path, style="dim")
            info_panel.update(feedback_text)
            
            # Reset to normal info after delay
            def reset_info():
                if self.current_selection:
                    path = self.current_selection["path"]
                    file_type = self.current_selection["type"]
                    size = None
                    if file_type == "file":
                        try:
                            size = path.stat().st_size
                        except (OSError, PermissionError):
                            pass
                    info_panel.update_info(path, file_type, size)
            
            self.set_timer(2.0, reset_info)
            
        except (subprocess.CalledProcessError, FileNotFoundError):
            # If clipboard copy fails, just ring the bell
            self.bell()


@click.command()
@click.argument('path', default='.', type=click.Path(exists=True))
def main(path: str):
    """
    vtree - A modern, interactive terminal-based file tree viewer
    
    PATH: Directory to display (default: current directory)
    """
    try:
        app = VTreeApp(path)
        app.run()
    except Exception as e:
        click.echo(f"Error: {e}", err=True)
        sys.exit(1)


if __name__ == "__main__":
    main()